@@ -26,19 +26,28 @@
  * vim: set sw=2 ts=2 et tw=80 : 
  */
 
-var convert = require('./lib/convert'),
+var Buffer = require('buffer').Buffer,
     crypto = require('crypto'),
-    http = require('http'),
-    https = require('https'),
+    http = require('./http');
     querystring = require('querystring'),
     url = require('url'),
     xrds = require('./lib/xrds');
 
 var _associations = {};
 var _discoveries = {};
+var _nonces = {};
+
+var AX_MAX_VALUES_COUNT = 1000;
 
 var openid = exports;
 
+var _get = http.get;
+var _post = http.post;
+
+function hasOwnProperty(obj, prop) {
+  return Object.prototype.hasOwnProperty.call(obj, prop);
+}
+
 openid.RelyingParty = function(returnUrl, realm, stateless, strict, extensions)
 {
   this.returnUrl = returnUrl;
@@ -56,7 +65,28 @@ openid.RelyingParty.prototype.authenticate = function(identifier, immediate, cal
 
 openid.RelyingParty.prototype.verifyAssertion = function(requestOrUrl, callback)
 {
-  openid.verifyAssertion(requestOrUrl, callback, this.stateless, this.extensions, this.strict);
+  openid.verifyAssertion(requestOrUrl, this.returnUrl, callback, this.stateless, this.extensions, this.strict);
+}
+
+
+
+var _btwoc = function(i)
+{
+  if(i.charCodeAt(0) > 127)
+  {
+    return String.fromCharCode(0) + i;
+  }
+  return i;
+}
+
+var _unbtwoc = function(i)
+{
+  if(i[0] === String.fromCharCode(0))
+  {
+    return i.substr(1);
+  }
+
+  return i;
 }
 
 var _isDef = function(e)
@@ -65,14 +95,43 @@ var _isDef = function(e)
   return e !== undefined;
 }
 
-var _toBase64 = function(binary)
+// Find the most up-to-date and usable way to create buffers
+var _buffer = null;
+if (typeof(Buffer.from) === 'function') {
+  // Some older Node versions throw an exception when 
+  // buffers with binary encoding are created using the
+  // from function, so if that happens we have to resort
+  // to constructor based creation.
+  try {
+    Buffer.from('openid', 'binary');
+    _buffer = Buffer.from;
+  }
+  catch(_) {
+  }
+}
+if (_buffer === null) {
+  // Either the Node version is too old to have a Buffer.from,
+  // or the Buffer.from call failed with binary encoding.
+  // Either way, use the (deprecated from node v6) constructor.
+  _buffer = function(str, enc) { return new Buffer(str, enc); };
+} 
+
+var _base64encode = function(str) {
+  return _buffer(str, 'binary').toString('base64');
+};
+
+var _base64decode = function(str) {
+  return _buffer(str, 'base64').toString('binary');
+};
+
+var _bigIntToBase64 = function(binary)
 {
-  return convert.base64.encode(convert.btwoc(binary));
+  return _base64encode(_btwoc(binary));  
 }
 
-var _fromBase64 = function(str)
+var _bigIntFromBase64 = function(str)
 {
-  return convert.unbtwoc(convert.base64.decode(str));
+  return _unbtwoc(_base64decode(str));
 }
 
 var _xor = function(a, b)
@@ -148,7 +207,7 @@ var _buildUrl = function(theUrl, params)
     {
       for(var key in params)
       {
-        if(params.hasOwnProperty(key))
+        if(hasOwnProperty(params, key))
         {
           theUrl.query[key] = params[key];
         }
@@ -158,203 +217,27 @@ var _buildUrl = function(theUrl, params)
 
   return url.format(theUrl);
 }
-
-var _proxyRequest = function(protocol, options)
-{
-  /* 
-  If process.env['HTTP_PROXY_HOST'] and the env variable `HTTP_PROXY_POST`
-  are set, make sure path and the header Host are set to target url.
-
-  Similarly, `HTTPS_PROXY_HOST` and `HTTPS_PROXY_PORT` can be used
-  to proxy HTTPS traffic.
-
-  Proxies Example:
-      export HTTP_PROXY_HOST=localhost
-      export HTTP_PROXY_PORT=8080
-      export HTTPS_PROXY_HOST=localhost
-      export HTTPS_PROXY_PORT=8442
-
-  Function returns protocol which should be used for network request, one of
-  http: or https:
-  */
-  var targetHost = options.host;
-  var newProtocol = protocol;
-  if (!targetHost) return;
-  var updateOptions = function (envPrefix) {
-    var proxyHostname = process.env[envPrefix + '_PROXY_HOST'].trim();
-    var proxyPort = parseInt(process.env[envPrefix + '_PROXY_PORT'], 10);
-    if (proxyHostname.length > 0 && ! isNaN(proxyPort)) {
-
-      if (! options.headers) options.headers = {};
-
-      var targetHostAndPort = targetHost + ':' + options.port;
-
-      options.host = proxyHostname;
-      options.port = proxyPort;
-      options.path = protocol + '//' + targetHostAndPort + options.path;
-      options.headers['Host'] = targetHostAndPort;
-    }
-  };
-  if ('https:' === protocol &&
-      !! process.env['HTTPS_PROXY_HOST'] &&
-      !! process.env['HTTPS_PROXY_PORT']) {
-    updateOptions('HTTPS');
-    // Proxy server request must be done via http... it is responsible for
-    // Making the https request...    
-    newProtocol = 'http:';
-  } else if (!! process.env['HTTP_PROXY_HOST'] &&
-             !! process.env['HTTP_PROXY_PORT']) {
-    updateOptions('HTTP');
-  }
-  return newProtocol;
-}
-
-var _get = function(getUrl, params, callback, redirects)
-{
-  redirects = redirects || 5;
-  getUrl = url.parse(_buildUrl(getUrl, params));
-
-  var path = getUrl.pathname || '/';
-  if(getUrl.query)
-  {
-    path += '?' + getUrl.query;
-  }
-  var options = 
-  {
-    host: getUrl.hostname,
-    port: _isDef(getUrl.port) ? parseInt(getUrl.port, 10) :
-      (getUrl.protocol == 'https:' ? 443 : 80),
-    headers: { 'Accept' : 'application/xrds+xml,text/html,text/plain,*/*' },
-    path: path
-  };
-
-  var protocol = _proxyRequest(getUrl.protocol, options);
-
-  (protocol == 'https:' ? https : http).get(options, function(res)
-  {
-    var data = '';
-    res.on('data', function(chunk)
-    {
-      data += chunk;
-    });
-
-    var isDone = false;
-    var done = function()
-    {
-      if (isDone) return;
-      isDone = true;
-
-      if(res.headers.location && --redirects)
-      {
-        var redirectUrl = res.headers.location;
-        if(redirectUrl.indexOf('http') !== 0)
-        {
-          redirectUrl = getUrl.protocol + '//' + getUrl.hostname + ':' + options.port + (redirectUrl.indexOf('/') === 0 ? redirectUrl : '/' + redirectUrl);
-        }
-        _get(redirectUrl, params, callback, redirects);
-      }
-      else
-      {
-        callback(data, res.headers, res.statusCode);
-      }
-    }
-
-    res.on('end', function() { done(); });
-    res.on('close', function() { done(); });
-  }).on('error', function(error) 
-  {
-    return callback(error);
-  });
-}
-
-var _post = function(postUrl, data, callback, redirects)
-{
-  redirects = redirects || 5;
-  postUrl = url.parse(postUrl);
-
-  var path = postUrl.pathname || '/';
-  if(postUrl.query)
-  {
-    path += '?' + postUrl.query;
-  }
-
-  var encodedData = _encodePostData(data);
-  var options = 
-  {
-    host: postUrl.hostname,
-    path: path,
-    port: _isDef(postUrl.port) ? postUrl.port :
-      (postUrl.protocol == 'https:' ? 443 : 80),
-    headers: 
-    {
-      'Content-Type': 'application/x-www-form-urlencoded',
-      'Content-Length': encodedData.length
-    },
-    method: 'POST'
-  };
-
-  var protocol = _proxyRequest(postUrl.protocol, options);
-
-  (protocol == 'https:' ? https : http).request(options, function(res)
-  {
-    var data = '';
-    res.on('data', function(chunk)
-    {
-      data += chunk;
-    });
-
-    var isDone = false;
-    var done = function()
-    {
-      if (isDone) return;
-      isDone = true;
-
-      if(res.headers.location && --redirects)
-      {
-        _post(res.headers.location, data, callback, redirects);
-      }
-      else
-      {
-        callback(data, res.headers, res.statusCode);
-      }
-    }
-
-    res.on('end', function() { done(); });
-    res.on('close', function() { done(); });
-  }).on('error', function(error)
-  {
-    return callback(error);
-  }).end(encodedData);
-}
-
-var _encodePostData = function(data)
-{
-  var encoded = querystring.stringify(data);
-  return encoded;
-}
-
 var _decodePostData = function(data)
 {
   var lines = data.split('\n');
   var result = {};
   for (var i = 0; i < lines.length ; i++) {
-    var line = lines[i];
-    if (line.length > 0 && line[line.length - 1] == '\r') {
+      var line = lines[i];
+      if (line.length > 0 && line[line.length - 1] == '\r') {
       line = line.substring(0, line.length - 1);
-    }
-    var colon = line.indexOf(':');
-    if (colon === -1)
-    {
+      }
+      var colon = line.indexOf(':');
+      if (colon === -1)
+      {
       continue;
-    }
-    var key = line.substr(0, line.indexOf(':'));
-    var value = line.substr(line.indexOf(':') + 1);
-    result[key] = value;
+      }
+      var key = line.substr(0, line.indexOf(':'));
+      var value = line.substr(line.indexOf(':') + 1);
+      result[key] = value;
   }
 
   return result;
 }
-
 var _normalizeIdentifier = function(identifier)
 {
   identifier = identifier.replace(/^\s+|\s+$/g, '');
@@ -497,15 +380,15 @@ var _parseHtml = function(htmlUrl, html, callback, hops)
 
 var _parseHostMeta = function(hostMeta, callback)
 {
-  var match = /^Link: <([^\n\r]+)>;/.exec(hostMeta);
-  if(match != null)
+  var match = /^Link: <([^\n\r]+?)>;/.exec(hostMeta);
+  if (match != null && match.length > 0)
   {
-    var xriUrl = match[0].slice(7,match.length - 4);
+    var xriUrl = match[1];
     _resolveXri(xriUrl, callback);
   }
   else
   {
-    callback(null)
+    callback(null);
   }
 }
 
@@ -547,7 +430,7 @@ var _resolveXri = function(xriUrl, callback, hops)
       var contentType = headers['content-type'];
       // text/xml is not compliant, but some hosting providers refuse header
       // changes, so text/xml is encountered
-      if(contentType.indexOf('application/xrds+xml') === 0 || contentType.indexOf('text/xml') === 0)
+      if(contentType && (contentType.indexOf('application/xrds+xml') === 0 || contentType.indexOf('text/xml') === 0))
       {
         return callback(_parseXrds(xriUrl, data));
       }
@@ -597,7 +480,7 @@ var _resolveHostMeta = function(identifier, strict, callback, fallBackToProxy)
   var hostMetaUrl;
   if(fallBackToProxy && !strict)
   {
-    hostMetaUrl = 'https://www.google.com/accounts/o8/.well-known/host-meta?hd=' + host.host
+    hostMetaUrl = 'https://www.google.com/accounts/o8/.well-known/host-meta?hd=' + host.host;
   }
   else
   {
@@ -626,7 +509,7 @@ var _resolveHostMeta = function(identifier, strict, callback, fallBackToProxy)
         //the response to hostMetaUrl was some other http/html resource.
         //Therefore fallback to the proxy if no providers are found.
         _parseHostMeta(data, function(providers){
-          if((providers == null || providers.length == 0) && !fallBackToProxy && !strict){
+          if((providers == null || providers.length == 0) && !fallBackToProxy && !strict) {
             _resolveHostMeta(identifier, strict, callback, true);
           }
           else{
@@ -711,9 +594,9 @@ openid.associate = function(provider, callback, strict, algorithm)
   if(algorithm.indexOf('no-encryption') === -1)
   {
     dh = _createDiffieHellmanKeyExchange(algorithm);
-    params['openid.dh_modulus'] = _toBase64(dh.getPrime('binary'));
-    params['openid.dh_gen'] = _toBase64(dh.getGenerator('binary'));
-    params['openid.dh_consumer_public'] = _toBase64(dh.getPublicKey('binary'));
+    params['openid.dh_modulus'] = _bigIntToBase64(dh.getPrime('binary'));
+    params['openid.dh_gen'] = _bigIntToBase64(dh.getGenerator('binary'));
+    params['openid.dh_consumer_public'] = _bigIntToBase64(dh.getPublicKey('binary'));
   }
 
   _post(provider.endpoint, params, function(data, headers, statusCode)
@@ -735,7 +618,7 @@ openid.associate = function(provider, callback, strict, algorithm)
     {
       if(algorithm == 'DH-SHA1')
       {
-        if(strict && url.protocol != 'https:')
+        if(strict && provider.endpoint.toLowerCase().indexOf('https:') !== 0)
         {
           return callback({ message: 'Channel is insecure and no encryption method is supported by provider' }, null);
         }
@@ -746,7 +629,7 @@ openid.associate = function(provider, callback, strict, algorithm)
       }
       else if(algorithm == 'no-encryption-256')
       {
-        if(strict && url.protocol != 'https:')
+        if(strict && provider.endpoint.toLowerCase().indexOf('https:') !== 0)
         {
           return callback('Channel is insecure and no encryption method is supported by provider', null);
         }
@@ -789,13 +672,13 @@ openid.associate = function(provider, callback, strict, algorithm)
       }
       else
       {
-        var serverPublic = _fromBase64(data.dh_server_public);
-        var sharedSecret = convert.btwoc(dh.computeSecret(serverPublic, 'binary', 'binary'));
+        var serverPublic = _bigIntFromBase64(data.dh_server_public);
+        var sharedSecret = _btwoc(dh.computeSecret(serverPublic, 'binary', 'binary'));
         var hash = crypto.createHash(hashAlgorithm);
-        hash.update(sharedSecret);
+        hash.update(_buffer(sharedSecret, 'binary'));
         sharedSecret = hash.digest('binary');
-        var encMacKey = convert.base64.decode(data.enc_mac_key);
-        secret = convert.base64.encode(_xor(encMacKey, sharedSecret));
+        var encMacKey = _base64decode(data.enc_mac_key);
+        secret = _base64encode(_xor(encMacKey, sharedSecret));
       }
 
       if (!_isDef(data.assoc_handle)) {
@@ -876,7 +759,7 @@ openid.authenticate = function(identifier, returnUrl, realm, immediate, stateles
 
     var providerIndex = -1;
 
-    var chooseProvider = function successOrNext(error, authUrl)
+    (function chooseProvider(error, authUrl)
     {
       if(!error && authUrl)
       {
@@ -900,8 +783,9 @@ openid.authenticate = function(identifier, returnUrl, realm, immediate, stateles
         {
           return callback(null, authUrl);
         }
-        else {
-          successOrNext({ message: 'OpenID 1.0/1.1 provider cannot be used without a claimed identifier' });
+        else
+        {
+          chooseProvider({ message: 'OpenID 1.0/1.1 provider cannot be used without a claimed identifier' });
         }
       }
       if(++providerIndex >= providers.length)
@@ -913,7 +797,7 @@ openid.authenticate = function(identifier, returnUrl, realm, immediate, stateles
       if(stateless)
       {
         _requestAuthentication(currentProvider, null, returnUrl, 
-          realm, immediate, extensions || {}, successOrNext);
+          realm, immediate, extensions || {}, chooseProvider);
       }
 
       else
@@ -922,19 +806,17 @@ openid.authenticate = function(identifier, returnUrl, realm, immediate, stateles
         {
           if(error || !answer || answer.error)
           {
-            successOrNext(error || answer.error, null);
+            chooseProvider(error || answer.error, null);
           }
           else
           {
             _requestAuthentication(currentProvider, answer.assoc_handle, returnUrl, 
-              realm, immediate, extensions || {}, successOrNext);
+              realm, immediate, extensions || {}, chooseProvider);
           }
         });
         
       }
-    };
-
-    chooseProvider();
+    })();
   });
 }
 
@@ -951,7 +833,7 @@ var _requestAuthentication = function(provider, assoc_handle, returnUrl, realm,
 
   for (var i in extensions)
   {
-    if(!extensions.hasOwnProperty(i))
+    if(!hasOwnProperty(extensions, i))
     {
       continue;
     }
@@ -959,7 +841,7 @@ var _requestAuthentication = function(provider, assoc_handle, returnUrl, realm,
     var extension = extensions[i];
     for (var key in extension.requestParams)
     {
-      if (!extension.requestParams.hasOwnProperty(key)) { continue; }
+      if (!hasOwnProperty(extension.requestParams, key)) { continue; }
       params[key] = extension.requestParams[key];
     }
   }
@@ -1014,14 +896,14 @@ var _requestAuthentication = function(provider, assoc_handle, returnUrl, realm,
   callback(null, _buildUrl(provider.endpoint, params));
 }
 
-openid.verifyAssertion = function(requestOrUrl, callback, stateless, extensions, strict)
+openid.verifyAssertion = function(requestOrUrl, originalReturnUrl, callback, stateless, extensions, strict)
 {
   extensions = extensions || {};
   var assertionUrl = requestOrUrl;
   if(typeof(requestOrUrl) !== typeof(''))
   {
-    if(requestOrUrl.method == 'POST') {
-      if(requestOrUrl.headers['content-type'] == 'application/x-www-form-urlencoded') {
+    if(requestOrUrl.method.toUpperCase() == 'POST') {
+      if((requestOrUrl.headers['content-type'] || '').toLowerCase().indexOf('application/x-www-form-urlencoded') === 0) {
         // POST response received
         var data = '';
         
@@ -1040,7 +922,7 @@ openid.verifyAssertion = function(requestOrUrl, callback, stateless, extensions,
       
       return; // Avoid falling through to GET method assertion
     }
-    else if(requestOrUrl.method != 'GET') {
+    else if(requestOrUrl.method.toUpperCase() != 'GET') {
       return callback({ message: 'Invalid request method from OpenID provider' });
     }
     assertionUrl = requestOrUrl.url;
@@ -1049,9 +931,44 @@ openid.verifyAssertion = function(requestOrUrl, callback, stateless, extensions,
   assertionUrl = url.parse(assertionUrl, true);
   var params = assertionUrl.query;
 
+  if (!_verifyReturnUrl(assertionUrl, originalReturnUrl)) {
+      return callback({ message: 'Invalid return URL' });
+  }
   return _verifyAssertionData(params, callback, stateless, extensions, strict);
 }
 
+var _verifyReturnUrl = function (assertionUrl, originalReturnUrl) {
+  var receivedReturnUrl = assertionUrl.query['openid.return_to'];
+  if (!_isDef(receivedReturnUrl)) {
+    return false;
+  }
+
+  receivedReturnUrl = url.parse(receivedReturnUrl, true);
+  if (!receivedReturnUrl) {
+    return false;
+  }
+  originalReturnUrl = url.parse(originalReturnUrl, true);
+  if (!originalReturnUrl) {
+    return false;
+  }
+
+  if (originalReturnUrl.protocol !== receivedReturnUrl.protocol || // Verify scheme against original return URL
+      originalReturnUrl.host !== receivedReturnUrl.host || // Verify authority against original return URL
+      originalReturnUrl.pathname !== receivedReturnUrl.pathname) { // Verify path against current request URL
+    return false;
+  }
+
+  // Any query parameters that are present in the "openid.return_to" URL MUST also be present 
+  // with the same values in the URL of the HTTP request the RP received
+  for (var param in receivedReturnUrl.query) {
+    if (hasOwnProperty(receivedReturnUrl.query, param) && receivedReturnUrl.query[param] !== assertionUrl.query[param]) {
+      return false;
+    }
+  }
+
+  return true;
+}
+
 var _verifyAssertionData = function(params, callback, stateless, extensions, strict) {
   var assertionError = _getAssertionError(params);
   if(assertionError)
@@ -1063,7 +980,10 @@ var _verifyAssertionData = function(params, callback, stateless, extensions, str
     return callback({ message: 'Unable to invalidate association handle'});
   }
 
-  // TODO: Check nonce if OpenID 2.0
+  if (!_checkNonce(params)) {
+      return callback({ message: 'Invalid or replayed nonce' });
+  }
+
   _verifyDiscoveredInformation(params, stateless, extensions, strict, function(error, result)
   {
     return callback(error, result);
@@ -1099,6 +1019,52 @@ var _invalidateAssociationHandleIfRequested = function(params)
   return true;
 }
 
+var _checkNonce = function (params) {
+  if (!_isDef(params['openid.ns'])) {
+    return true; // OpenID 1.1 has no nonce
+  }
+  if (!_isDef(params['openid.response_nonce'])) {
+    return false;
+  }
+
+  var nonce = params['openid.response_nonce'];
+  var timestampEnd = nonce.indexOf('Z');
+  if (timestampEnd == -1) {
+    return false;
+  }
+
+  // Check for valid timestamp in nonce
+  var timestamp = new Date(Date.parse(nonce.substring(0, timestampEnd + 1)));
+  if (Object.prototype.toString.call(timestamp) !== '[object Date]' || isNaN(timestamp)) {
+    return false;
+  }
+    
+  // Remove old nonces from our store (nonces that are more skewed than 5 minutes)
+  _removeOldNonces();
+
+  // Check if nonce is skewed by more than 5 minutes
+  if (Math.abs(new Date().getTime() - timestamp.getTime()) > 300000) {
+    return false;
+  }
+
+  // Check if nonce is replayed
+  if (_isDef(_nonces[nonce])) {
+    return false;
+  }
+
+  // Store the nonce
+  _nonces[nonce] = timestamp;
+  return true;
+}
+
+var _removeOldNonces = function () {
+  for (var nonce in _nonces) {
+    if (hasOwnProperty(_nonces, nonce) && Math.abs(new Date().getTime() - _nonces[nonce].getTime()) > 300000) {
+      delete _nonces[nonce];
+    }
+  }
+}
+
 var _verifyDiscoveredInformation = function(params, stateless, extensions, strict, callback)
 {
   var claimedIdentifier = params['openid.claimed_id'];
@@ -1161,12 +1127,12 @@ var _verifyAssertionAgainstProviders = function(providers, params, stateless, ex
   for(var i = 0; i < providers.length; ++i)
   {
     var provider = providers[i];
-    if(!provider.version || provider.version.indexOf(params['openid.ns']) !== 0)
+    if(!!params['openid.ns'] && (!provider.version || provider.version.indexOf(params['openid.ns']) !== 0))
     {
       continue;
     }
 
-    if(provider.version.indexOf('2.0') !== -1)
+    if(!!provider.version && provider.version.indexOf('2.0') !== -1)
     {
       var endpoint = params['openid.op_endpoint'];
       if (provider.endpoint != endpoint) 
@@ -1181,7 +1147,7 @@ var _verifyAssertionAgainstProviders = function(providers, params, stateless, ex
       }
     }
 
-    if(provider.localIdentifier && provider.localIdentifier != params['openid.identity'])
+    if(!!provider.localIdentifier && provider.localIdentifier != params['openid.identity'])
     {
       return callback({ message: 'Identity in assertion response does not match discovered local identifier' });
     }
@@ -1196,7 +1162,7 @@ var _verifyAssertionAgainstProviders = function(providers, params, stateless, ex
       {
         for(var ext in extensions)
         {
-          if (!extensions.hasOwnProperty(ext))
+          if (!hasOwnProperty(extensions, ext))
           { 
             continue; 
           }
@@ -1264,7 +1230,7 @@ var _checkSignatureUsingAssociation = function(params, callback)
       message += param + ':' + value + '\n';
     }
 
-    var hmac = crypto.createHmac(association.type, convert.base64.decode(association.secret));
+    var hmac = crypto.createHmac(association.type, _buffer(association.secret, 'base64'));
     hmac.update(message, 'utf8');
     var ourSignature = hmac.digest('base64');
 
@@ -1287,7 +1253,7 @@ var _checkSignatureUsingProvider = function(params, provider, callback)
   };
   for(var key in params)
   {
-    if(params.hasOwnProperty(key) && key != 'openid.mode')
+    if(hasOwnProperty(params, key) && key != 'openid.mode')
     {
       requestParams[key] = params[key];
     }
@@ -1423,7 +1389,6 @@ openid.UserInterface.prototype.fillResult = function(params, result)
  *  - http://www.axschema.org/types/ 
  *  - http://code.google.com/intl/en-US/apis/accounts/docs/OpenID.html#Parameters
  */
-// TODO: count handling
 
 var attributeMapping = 
 {
@@ -1445,7 +1410,7 @@ openid.AttributeExchange = function AttributeExchange(options)
   var optional = [];
   for (var ns in options)
   {
-    if (!options.hasOwnProperty(ns)) { continue; }
+    if (!hasOwnProperty(options, ns)) { continue; }
     if (options[ns] == 'required')
     {
       required.push(ns);
@@ -1481,12 +1446,13 @@ openid.AttributeExchange = function AttributeExchange(options)
 openid.AttributeExchange.prototype.fillResult = function(params, result)
 {
   var extension = _getExtensionAlias(params, 'http://openid.net/srv/ax/1.0') || 'ax';
-  var regex = new RegExp('^openid\\.' + extension + '\\.(value|type)\\.(\\w+)$');
+  var regex = new RegExp('^openid\\.' + extension + '\\.(value|type|count)\\.(\\w+)(\\.(\\d+)){0,1}$');
   var aliases = {};
+  var counters = {};
   var values = {};
   for (var k in params)
   {
-    if (!params.hasOwnProperty(k)) { continue; }
+    if (!hasOwnProperty(params, k)) { continue; }
     var matches = k.match(regex);
     if (!matches)
     {
@@ -1496,9 +1462,37 @@ openid.AttributeExchange.prototype.fillResult = function(params, result)
     {
       aliases[params[k]] = matches[2];
     }
+    else if (matches[1] == 'count')
+    {
+      //counter sanitization
+      var count = parseInt(params[k], 10);
+
+      // values number limitation (potential attack by overflow ?)
+      counters[matches[2]] = (count < AX_MAX_VALUES_COUNT) ? count : AX_MAX_VALUES_COUNT ;
+    }
     else
     {
-      values[matches[2]] = params[k];
+      if (matches[3])
+      {
+        //matches multi-value, aka "count" aliases
+
+        //counter sanitization
+        var count = parseInt(matches[4], 10);
+
+        // "in bounds" verification
+        if (count > 0 && count <= (counters[matches[2]] || AX_MAX_VALUES_COUNT))
+        {
+          if (!values[matches[2]]) {
+            values[matches[2]] = [];
+          }
+          values[matches[2]][count-1] = params[k];
+        }
+      }
+      else
+      {
+        //matches single-value aliases
+        values[matches[2]] = params[k];
+      }
     }
   }
   for (var ns in aliases) 
@@ -1602,7 +1596,7 @@ openid.PAPE.prototype.fillResult = function(params, result)
   var paramString = 'openid.' + extension + '.';
   var thisParam;
   for (var p in params) {
-    if (params.hasOwnProperty(p)) {
+    if (hasOwnProperty(params, p)) {
       if (p.substr(0, paramString.length) === paramString) {
         thisParam = p.substr(paramString.length);
         if (thisParam === 'auth_policies') {
